/**
 * Copyright (c) 2008 Mia-Software. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License v1.0 
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributor: 
 *     Fabien Giquel
 */

package org.eclipse.gmt.modisco.discoverer.kdm.source;

import java.io.File;

import org.eclipse.core.resources.IContainer;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl;
import org.eclipse.gmt.modisco.omg.kdm.source.Directory;
import org.eclipse.gmt.modisco.omg.kdm.source.InventoryContainer;
import org.eclipse.gmt.modisco.omg.kdm.source.InventoryItem;
import org.eclipse.gmt.modisco.omg.kdm.source.InventoryModel;
import org.eclipse.gmt.modisco.omg.kdm.source.impl.SourceFactoryImpl;
import org.eclipse.jdt.core.IJavaProject;

/**
 * Some services for building a basic file system representation, based on 
 * Source sub package from KDM metamodel.
 * 
 */
public class KDMSourceDiscoverer {

	private boolean rootIsProject = false;
	private InventoryContainer rootContainer = null;

	/**
	 * Create a new File system representation for the given File
	 * 
	 * @param aFile
	 *            must be a directory
	 * @return
	 */
	public Resource getKDMModelFromDirectory(File aFile) {
		if (!aFile.isDirectory())
			return null;
		this.rootIsProject = false;
		this.rootContainer = null;
		return createKDMModelFromFile(aFile, aFile.getName());
	}

	/**
	 * Create a new File system representation for the given Eclipse Java
	 * project
	 * 
	 * @param aProject
	 * @return
	 */
	public Resource getKDMModelFromJavaProject(IJavaProject aProject) {
		this.rootIsProject = true;
		this.rootContainer = null;
		return createKDMModelFromFile(aProject.getResource().getLocation()
				.toFile(), aProject.getElementName());
	}

	/**
	 * Create a new File system representation for the given Eclipse container
	 * 
	 * @param aContainer
	 * @return
	 */
	public Resource getKDMModelFromContainer(IContainer aContainer) {
		this.rootIsProject = (aContainer == aContainer.getProject());
		this.rootContainer = null;
		return createKDMModelFromFile(aContainer.getLocation().toFile(),
				aContainer.getName());
	}

	private Resource createKDMModelFromFile(File directoryFile, String name) {
		Resource resultModel = new XMIResourceImpl();
		InventoryModel inventoryModel = SourceFactoryImpl.eINSTANCE
				.createInventoryModel();
		inventoryModel.setName(name);
		resultModel.getContents().add(inventoryModel);

		InventoryContainer project = createFileSystemTree(directoryFile);
		inventoryModel.getInventoryElement().add(project);
		return resultModel;
	}

	private InventoryContainer createFileSystemTree(File directory) {
		InventoryContainer resultDirectory;
		if (this.rootContainer == null && this.rootIsProject)
			resultDirectory = this.rootContainer = SourceFactoryImpl.eINSTANCE
					.createProject();
		else {
			resultDirectory = SourceFactoryImpl.eINSTANCE.createDirectory();
			((Directory) resultDirectory).setPath(directory.getAbsolutePath());
			if (this.rootContainer == null)
				this.rootContainer = resultDirectory;
		}
		resultDirectory.setName(directory.getName());

		File[] children = directory.listFiles();
		for (File child : children) {
			if (child.isFile()) {
				InventoryItem childItem = SourceFactoryImpl.eINSTANCE
						.createInventoryItem();
				childItem.setPath(child.getAbsolutePath());
				childItem.setName(child.getName());
				resultDirectory.getInventoryElement().add(childItem);
			} else {
				Directory childDirectory = (Directory) createFileSystemTree(child);
				resultDirectory.getInventoryElement().add(childDirectory);
			}
		}
		return resultDirectory;
	}

}
